# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
from typing import Optional, Union
from datasets import load_dataset, concatenate_datasets
from datasets import disable_progress_bars
import re
import numpy as np
from copy import deepcopy

"""This module uses the _dataset_default_params dict to control what datasets load_example_dataset can load and how they are loaded.

You can add new datasets to the loader by adding a new entry in the _dataset_default_params dict like:

'my_dataset_name': {

    'load_params': {'path': 'hf_hub_org_name/org_dataset_name', # HF dataset ref
                    'name': 'if_hf_dataset_has_multi_files', # needed for specific datasets
                    'split': 'train'}, # optional
    'extra_processing': {} # optional, see examples in _dataset_default_params dict

}
"""
_dataset_default_params = {
    "ai2_arc": {
        "load_params": {
            "path": "allenai/ai2_arc",  # HF Hub dataset name
            "name": "ARC-Easy",  # HF Hub filename
            "split": "test",
        },  # HF Hub dataset split
        "extra_processing": {
            "rename_columns": {
                "answerKey": "answer"
            },  # renaming is always the first operation
            "strip_whitespace": [
                "answer"
            ],  # notice we're referencing the renamed col name
            "to_upper": ["answer"],
            "combine_question_and_choices": {
                "question_col": "question",
                "choice_col": "choices",
                "choice_text_col": "text",
                "choice_label_col": "label",
            },
            "subset_columns": ["question", "answer"],
        },
    },
    "csqa": {
        "load_params": {"path": "skrishna/CSQA_preprocessed", "split": "train"},
        "extra_processing": {
            "rename_columns": {
                "question": "q_only",
                "answerKey": "answer",
                "inputs": "question",
            },
            "to_upper": ["answer"],
            "subset_columns": ["question", "answer"],
        },
    },
    "gsm8k": {
        "load_params": {"path": "openai/gsm8k", "name": "main", "split": "train"},
        "extra_processing": {
            "regex_filters": [
                {
                    "pattern": "#### ([-+]?\d*\\.\d+|[-+]?\d+)",  # regex pattern
                    "col": "answer",  # dataset col to apply pattern to
                    "operation": "search",  # type of regex operation
                    "group": 1,
                }
            ],  # capture group desired
            "subset_columns": ["question", "answer"],
        },
    },
    "nq_open": {
        "load_params": {
            "path": "google-research-datasets/nq_open",
            "split": "validation",
        },
        "extra_processing": {
            "to_lower": ["answer"],
            "subset_columns": ["question", "answer"],
        },
    },
    "popqa": {
        "load_params": {"path": "akariasai/PopQA", "split": "test"},
        "extra_processing": {
            "rename_columns": {"possible_answers": "answer"},
            "to_lower": ["answer"],
            "subset_columns": ["question", "answer"],
        },
    },
    "svamp": {
        "load_params": {"path": "Chilled/SVAMP"},
        "extra_processing": {
            "concat": "all",
            "rename_columns": {"question_concat": "question", "Answer": "answer"},
            "regex_filters": [
                {
                    "pattern": "#### ([-+]?\d*\\.\d+|[-+]?\d+)",  # get numbers only (like gsm8k)
                    "col": "answer",
                    "operation": "search",
                    "group": 1,
                }
            ],
            "subset_columns": ["question", "answer"],
        },
    },
}


def list_dataset_names() -> list:
    """
    List all available example dataset names in uqlm.

    Returns
    -------
    list
        A list of available datasets.

    Example
    -------
    >>> from uqlm.utils.dataloader import list_dataset_names
    >>> list_dataset_names()
    ['ai2_arc', 'csqa', 'dialogue_sum', 'gsm8k', 'nq_open', 'popqa', 'svamp', 'triviaqa']
    """
    return list(_dataset_default_params.keys())


def load_example_dataset(
    name: str, n: int = None, cols: Optional[Union[list, str]] = None
) -> pd.DataFrame:
    """
    Load a dataset for testing purposes.

    Parameters
    ----------
    name : str
        The name of the dataset to load. Must be one of "svamp", "gsm8k", "ai2_arc",
        "csqa", "nq_open", "popqa"
        
    n : int, optional
        Number of rows to load from the dataset.

    Returns
    -------
    pd.DataFrame
        The loaded dataset.

    Example
    -------
    >>> from uqlm.utils.dataloader import load_example_dataset
    >>> df = load_example_dataset("gsm8k", n=1000)
    >>> df.shape
    (1000, 2)
    """
    dataset_dict = deepcopy(_dataset_default_params)
    if name in dataset_dict.keys():  # loads from huggingface hub
        disable_progress_bars()  # disable hf tqdm bars b/c it's a little ugly
        print(f"Loading dataset - {name}...")
        ds = load_dataset(**dataset_dict[name]["load_params"])
        print("Processing dataset...")
        extras = dataset_dict[name].get("extra_processing", dict())
        if extras:
            if extras.get("concat"):  # combine different splits into one
                split = extras.get("concat")
                if split == "all":  # combine splits into one dataset
                    ds = concatenate_datasets([ds[s] for s in ds])
                extras.pop("concat")
        df = ds.to_pandas()
        if cols:
            extras["subset_columns"] = cols
        if extras:
            df = _dataset_processing(df=df, **extras)  # data wrangling on single df
        if isinstance(n, int):
            df = df.iloc[:n]
        print("Dataset ready!")
        return df
    else:
        raise FileNotFoundError(
            f"uqlm could not find the dataset '{name}'.\nPlease use `uqlm.utils.dataloader.list_dataset_names()` for available sample datasets."
        )


def _dataset_processing(
    df: pd.DataFrame,
    rename_columns: dict = None,
    subset_columns: list = None,
    to_upper: list = None,
    to_lower: list = None,
    combine_question_and_choices: dict = None,
    strip_non_numeric: list = None,
    strip_whitespace: list = None,
    regex_filters: list[dict] = None,
) -> pd.DataFrame:
    """
    Process a dataset with various operations.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe to process.
    rename_columns : dict, optional
        A dictionary mapping old column names to new column names.
    subset_columns : list, optional
        A list of columns to keep in the dataframe.
    to_upper : list, optional
        A list of columns whose string values should be converted to uppercase.
    to_lower: list, optional
        A list of columns whose string values should be converted to lowercase.
    combine_question_and_choices : dict, optional
        A dictionary with parameters to combine question and question choice columns.
    strip_non_numeric : list, optional
        A list of columns from which to strip non-numeric characters.
    strip_whitespace : list, optional
        A list of columns from which to strip whitespace characters.
    regex_filters: list[dict]
        A list of dictionaries like `{'pattern':r'', 'col':''}` to describe regex transformations to apply on the dataset.
    Returns
    -------
    pd.DataFrame
        The processed dataframe.

    Raises
    ------
    TypeError
        If the input `df` is not a pandas DataFrame.

    Example
    -------
    >>> df = pd.DataFrame({'A': ['a', 'b', 'c'], 'B': ['1', '2', '3']})
    >>> _dataset_processing(df, rename_columns={'A': 'a'}, to_upper=['a'])
       a  B
    0  A  1
    1  B  2
    2  C  3
    """

    if not isinstance(df, pd.DataFrame):
        raise TypeError(
            f"Dataset processing requires 'pd.DataFrame' but received '{type(df)}'"
        )

    if rename_columns:
        df = df.rename(columns=rename_columns)
    if strip_non_numeric:
        for col in strip_non_numeric:
            df[col] = df[col].apply(lambda x: "".join(c for c in x if c.isdigit()))
    if strip_whitespace:
        for col in strip_whitespace:
            df[col] = df[col].str.replace(" ", "")
    if to_upper:
        for col in to_upper:
            if isinstance(df[col][0], (list, np.ndarray)):
                df[col] = df[col].apply(lambda x: [s.upper() for s in x])
            else:
                df[col] = df[col].str.upper()
    if to_lower:
        for col in to_lower:
            if isinstance(df[col][0], (list, np.ndarray)):
                df[col] = df[col].apply(lambda x: [s.lower() for s in x])
            else:
                df[col] = df[col].str.lower()
    if combine_question_and_choices:
        df = _combine_question_and_choices(df, **combine_question_and_choices)
    if regex_filters:
        for rfilter in regex_filters:
            if rfilter["operation"] == "search":
                if not rfilter.get("group", None):
                    rfilter["group"] == 0
                df[rfilter["col"]] = df[rfilter["col"]].apply(
                    lambda x: re.search(rfilter["pattern"], x).group(rfilter["group"])
                    if re.search(rfilter["pattern"], x)
                    else x
                )
    if subset_columns:
        cols = subset_columns
        if isinstance(cols, (list, str)):
            if isinstance(cols, str):
                cols = [cols]
            cols_in_df = [x for x in cols if x in df.columns]
            df = df[cols_in_df]
            if df.shape[1] != len(cols):
                print("WARNING: some specified columns not found in dataset...")

    return df


def _combine_question_and_choices(
    df: pd.DataFrame,
    question_col: str,
    choice_col: Union[str, list] = None,
    choice_text_col: str = None,
    choice_label_col: str = None,
    save_original_question: bool = False,
) -> pd.DataFrame:
    """
    Combine question and choices columns in a dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        The dataframe to process.
    question_col : str
        The name of the question column.
    choice_col : Union[str, list], optional
        The name(s) of the choice column(s). If a string, it should be the name of a column containing dictionaries.
        If a list, it should be the names of columns containing choices.
    choice_text_col : str, optional
        The name of the choice text column, used when `choice_col` is a string.
    choice_label_col : str, optional
        The name of the choice label column, used when `choice_col` is a string.
    save_original_question : bool, optional
        Whether to save the original question column as 'original_question'.

    Returns
    -------
    pd.DataFrame
        The processed dataframe with combined question and choices.

    Raises
    ------
    TypeError
        If `choice_col` is not a string or a list.

    Example
    -------
    >>> df = pd.DataFrame({
    ...     'question': ['What is the capital of France?', 'What is 2+2?'],
    ...     'choices': [{'text': ['Paris', 'London'], 'label': ['A', 'B']}, {'text': ['3', '4'], 'label': ['A', 'B']}]
    ... })
    >>> _combine_question_and_choices(df, question_col='question', choice_col='choices', choice_text_col='text', choice_label_col='label')
    >>> df
                               question
    0  What is the capital of France? A) Paris B) London
    1                          What is 2+2? A) 3 B) 4
    """
    if save_original_question and question_col == "question":
        df["original_question"] = df["question"].copy()
    if isinstance(choice_col, str):
        if isinstance(
            df[choice_col][0], dict
        ):  # example of this format is allenai/ai2_arc
            df["question"] = (
                df[question_col]
                + " "
                + df[choice_col].apply(
                    lambda x: " ".join(
                        [
                            f"{label}) {text}"
                            for text, label in zip(
                                x[choice_text_col], x[choice_label_col]
                            )
                        ]
                    )
                )
            )
    elif isinstance(choice_col, list):
        # TODO when a dataset needs this... ex would be where cols are like answerA, answerB
        pass
    else:
        raise TypeError(
            f"'choice_col' must be str or list, but received '{type(choice_col)}'"
        )
    return df


# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from typing import List

from uqlm.black_box.baseclass.similarity_scorer import SimilarityScorer


class MatchScorer(SimilarityScorer):
    def __init__(self) -> None:
        """
        Class for computing exact match rates between original responses and candidates. This 
        method is based on Cole et al.(2023) :footcite:`cole2023selectivelyansweringambiguousquestions`.
        """
        pass

    def evaluate(
        self, responses: List[str], sampled_responses: List[List[str]]
    ) -> List[float]:
        """
        This method computes exact match rates for the provided pairs of texts.

        Parameters
        ----------
        responses : list of strings
            Original LLM response

        sampled_responses : list of list of strings
            Candidate responses to be compared to the original response

        Returns
        -------
        List of float
            Exact match rates
        """
        return [
            self._compute_score(response=o, candidates=mr)
            for o, mr in zip(responses, sampled_responses)
        ]

    @staticmethod
    def _compute_score(response: str, candidates: List[str]) -> List[float]:
        """Get mean exact match rate between response and set of candidates"""
        return np.mean([1 if response == c else 0 for c in candidates])


# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, List, Optional

from uqlm.scorers.baseclass.uncertainty import UncertaintyQuantifier, UQResult


class SemanticEntropy(UncertaintyQuantifier):
    def __init__(
        self,
        llm=None,
        postprocessor: Any = None,
        device: Any = None,
        use_best: bool = True,
        system_prompt: str = "You are a helpful assistant.",
        max_calls_per_min: Optional[int] = None,
        use_n_param: bool = False,
        sampling_temperature: float = 1.0,
        verbose: bool = False,
        nli_model_name: str = "microsoft/deberta-large-mnli",
        max_length: int = 2000,
        discrete: bool = True,
    ) -> None:
        """
        Class for computing Discrete Semantic Entropy-based confidence scores. For more on semantic entropy,
        refer to Farquhar et al.(2024) :footcite:`farquhar2024detectinghallucinations`.

        Parameters
        ----------
        llm : langchain `BaseChatModel`, default=None
            A langchain llm `BaseChatModel`. User is responsible for specifying temperature and other
            relevant parameters to the constructor of their `llm` object.

        postprocessor : callable, default=None
            A user-defined function that takes a string input and returns a string. Used for postprocessing
            outputs.

        device: str or torch.device input or torch.device object, default="cpu"
            Specifies the device that NLI model use for prediction. Only applies to 'semantic_negentropy', 'noncontradiction' 
            scorers. Pass a torch.device to leverage GPU.
            
        use_best : bool, default=True
            Specifies whether to swap the original response for the uncertainty-minimized response
            based on semantic entropy clusters.

        system_prompt : str or None, default="You are a helpful assistant."
            Optional argument for user to provide custom system prompt

        max_calls_per_min : int, default=None
            Specifies how many api calls to make per minute to avoid a rate limit error. By default, no
            limit is specified.
            
        sampling_temperature : float, default=1.0
            The 'temperature' parameter for llm model to generate sampled LLM responses. Must be greater than 0.

        use_n_param : bool, default=False
            Specifies whether to use `n` parameter for `BaseChatModel`. Not compatible with all
            `BaseChatModel` classes. If used, it speeds up the generation process substantially when num_responses > 1.

        verbose : bool, default=False
            Specifies whether to print the index of response currently being scored.

        nli_model_name : str, default="microsoft/deberta-large-mnli"
            Specifies which NLI model to use. Must be acceptable input to AutoTokenizer.from_pretrained() and 
            AutoModelForSequenceClassification.from_pretrained()
            
        max_length : int, default=2000
            Specifies the maximum allowed string length. Responses longer than this value will be truncated to 
            avoid OutOfMemoryError
        """
        super().__init__(
            llm=llm,
            device=device,
            system_prompt=system_prompt,
            max_calls_per_min=max_calls_per_min,
            use_n_param=use_n_param,
            postprocessor=postprocessor,
        )
        self.nli_model_name = nli_model_name
        self.max_length = max_length
        self.verbose = verbose
        self.use_best = use_best
        self.sampling_temperature = sampling_temperature
        self.prompts = None
        self._setup_nli(nli_model_name)
        self.nli_scorer.discrete = discrete
        
    async def generate_and_score(
        self,
        prompts: List[str],
        num_responses: int = 5,
    ) -> UQResult:
        """
        Evaluate discrete semantic entropy score on LLM responses for the provided prompts.

        Parameters
        ----------
        prompts : list of str
            A list of input prompts for the model.

        num_responses : int, default=5
            The number of sampled responses used to compute consistency.

        Returns
        -------
        UQResult
            UQResult, containing data (prompts, responses, and semantic entropy scores) and metadata
        """
        self.prompts = prompts
        self.num_responses = num_responses
        self.nli_scorer.num_responses = num_responses

        responses = await self.generate_original_responses(prompts)
        sampled_responses = await self.generate_candidate_responses(prompts)
        return self.score(
            responses = responses, sampled_responses = sampled_responses,
        )
    
    def score(
        self,
        responses: List[str] = None,
        sampled_responses: List[List[str]] = None,
    ) -> UQResult:
        """
        Evaluate discrete semantic entropy score on LLM responses for the provided prompts.

        Parameters
        ----------
        responses : list of str, default=None
            A list of model responses for the prompts. If not provided, responses will be generated with the provided LLM.

        sampled_responses : list of list of str, default=None
            A list of lists of sampled model responses for each prompt. These will be used to compute consistency scores by comparing to 
            the corresponding response from `responses`. If not provided, sampled_responses will be generated with the provided LLM.

        Returns
        -------
        UQResult
            UQResult, containing data (responses, sampled responses, and semantic entropy scores) and metadata
        """
        self.responses = responses 
        self.sampled_responses = sampled_responses
        self.num_responses = len(self.sampled_responses[0])
        self.nli_scorer.num_responses = self.num_responses

        n_prompts = len(self.responses)
        semantic_entropy = [None] * n_prompts
        best_responses = [None] * n_prompts

        print("Computing confidence scores...")
        for i in range(n_prompts):
            candidates = [self.responses[i]] + self.sampled_responses[i]
            tmp = self.nli_scorer._semantic_entropy_process(candidates=candidates, i=i)
            best_responses[i], semantic_entropy[i], scores = tmp

        confidence_scores = [
            1 - ne for ne in self.nli_scorer._normalize_entropy(semantic_entropy)
        ]

        result = {
            "data": {
                "responses": best_responses if self.use_best else self.responses,
                "entropy_values": semantic_entropy,
                "confidence_scores": confidence_scores,
                "sampled_responses": self.sampled_responses,
            },
            "metadata": {
                "parameters": {
                    "temperature": None if not self.llm else self.llm.temperature,
                    "sampling_temperature": None
                    if not self.sampling_temperature
                    else self.sampling_temperature,
                    "num_responses": self.num_responses,
                },
            },
        }
        if self.prompts:
            result["data"]["prompts"] = self.prompts
        return UQResult(result)


# Copyright 2025 CVS Health and/or one of its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import contextlib
import io

import numpy as np
import pandas as pd
from typing import Any, Dict, List, Optional

from uqlm.utils.response_generator import ResponseGenerator


KEYWORDS_TO_SCORES_DICT = {
    round(0.0, 1): ["incorrect", "not correct", "not right", "wrong"],
    0.5: ["not sure", "not certain", "unsure", "uncertain"],
    1.0: ["correct", "right"],
}

CHOICES_2_CLASS = """\"Correct\", \"Incorrect\""""
CHOICES_3_CLASS = CHOICES_2_CLASS + """, or \"I am not sure\""""

CONTINUOUS_SCORE_INSTRUCTION = """
How likely is the above answer to be correct? Analyze the answer and give your confidence in this answer between 0 (lowest) and 100 (highest), with 100 being certain the answer is correct, and 0 being certain the answer is incorrect. THE CONFIDENCE RATING YOU PROVIDE MUST BE BETWEEN 0 and 100. ONLY RETURN YOUR NUMERICAL SCORE WITH NO SURROUNDING TEXT OR EXPLANATION.

# Example 1
## Data to analyze
Question: Who was the first president of the United States?, Proposed Answer: Benjamin Franklin.

## Your response
4 (highly certain the proposed answer is incorrect)

# Example 2
## Data to analyze
Question: What is 2+2?, Proposed Answer: 4

## Your response
99 (highly certain the proposed answer is correct)
"""


def cat_instruction(choices):
    return f"""Your task is to look at the question and answer provided and determine if the answer is correct. You are to respond with ONLY one of: {choices}. YOUR ANSWER MUST ONLY CONTAIN ONE OF {choices}. DO NOT ANSWER THE QUESTION AGAIN. ONLY DETERMINE IF THE ANSWER TO THE QUESTION IS {choices}."""


TEMPLATE_TO_INSTRUCTION = {
    "continuous": CONTINUOUS_SCORE_INSTRUCTION,
    "true_false_uncertain": cat_instruction(CHOICES_3_CLASS),
    "true_false": cat_instruction(CHOICES_2_CLASS),
}


class LLMJudge(ResponseGenerator):
    def __init__(
        self,
        llm: Any,
        max_calls_per_min: Optional[int] = None,
        scoring_template: str = "true_false_uncertain",
        system_prompt: Optional[str] = None,
        template_ques_ans: Optional[str] = None,
        keywords_to_scores_dict: Optional[Dict] = None,
    ) -> None:
        """
        Class for using LLM-as-a-judge to score proposed answers to questions based on correctness. Three off-the-shelf
        templates are offered: incorrect/uncertain/correct (0/0.5/1), incorrect/correct (0/1), and continuous score (0 to 1).
        Customization is also supported for user-provided classification-based judging templates. The correct/incorrect/uncertain
        template is based on Chen and Mueller(2023) :footcite:`chen2023quantifyinguncertaintyanswerslanguage`

        Parameters
        ----------
        llm : langchain llm object
            A langchain llm object to get passed to chain constructor. User is responsible for specifying
            temperature and other relevant parameters to the constructor of their `llm` object.

        max_calls_per_min : int, default=None
            Specifies how many api calls to make per minute to avoid a rate limit error. By default, no
            limit is specified.

        scoring_template : {'true_false_uncertain', 'true_false', 'continuous'}, default='true_false_uncertain'
             specifies which off-the-shelf template to use, if any. Three off-the-shelf templates offered:
             incorrect/uncertain/correct (0/0.5/1), incorrect/correct (0/1), and continuous score (0 to 1).
             These templates are respectively specified as 'true_false_uncertain', 'true_false', and 'continuous'

        system_prompt : str or None, default=None
            Optional argument for user to provide custom system prompt. If None, a default instruction
            system prompt will be used.

        template_ques_ans : f-string, default=None
            Template for self reflection question, which includes question and answer to
            compute LLM judge score. Use this to define the LLM response format, if required update
            argument "keywords_to_scores_dict" accordingly. Must be formatted so that template_ques_ans.format(question, answer)
            places question and answer appropriately in the string. Defaults to variation of Chen et al. (2023).

        keywords_to_scores_dict : dict, default=None
            Keys must be scores, values must be list of strings containing keywords to search. If None, the default
            dictionary will be used: {
            0.0: ["incorrect", "not correct", "not right"],
            0.5: ["not sure", "not certain", "unsure", "uncertain"],
            1.0: ["correct", "right"],
            }
        """
        super().__init__(llm=llm, max_calls_per_min=max_calls_per_min)
        self.scoring_template = scoring_template
        self.template_ques_ans = template_ques_ans
        self.keywords_to_scores_dict = keywords_to_scores_dict
        self._validate_inputs()
        self.system_prompt = self.instruction if not system_prompt else system_prompt

    async def judge_responses(
        self, prompts: List[str], responses: List[str], retries: int = 5
    ) -> Dict[str, Any]:
        """
        Judge responses for correctness.
        
        Parameters
        ----------
        prompts : list of str
            A list of input prompts for the model.

        responses: list of str
            A list of model responses for the provided prompts.

        retries : int, default=5
            Number of times to retry for failed score extraction

        Returns
        -------
        Dict
            Dictionary containing Q/A concatenation prompts, judge responses, and judge scores
        """
        concatenated_qa = [
            self.template_ques_ans.format(prompts[i], responses[i])
            for i in range(len(prompts))
        ]
        print("Generating LLMJudge scores...")
        with contextlib.redirect_stdout(io.StringIO()):
            data = await self.generate_responses(prompts=concatenated_qa, count=1)
        df = pd.DataFrame(
            {
                "judge_prompts": data["data"]["prompt"],
                "judge_responses": data["data"]["response"],
                "scores": self._extract_answers(responses=data["data"]["response"]),
            }
        )
        retry = 0
        while retry <= retries:
            retry += 1
            df_sub = df[pd.isna(df.scores)]
            if len(df_sub) > 0:
                with contextlib.redirect_stdout(io.StringIO()):
                    tmp = await self.generate_responses(
                        prompts=list(df_sub.judge_prompts), 
                        count=1, 
                        system_prompt=self.system_prompt
                    )
                df.loc[df_sub.index, "scores"] = self._extract_answers(
                    responses=tmp["data"]["response"]
                )
        return {col: list(df[col]) for col in df.columns}

    def _default_template_ques_ans(self):
        """Constructs default question-answer template"""
        qa_text = "Question: {}, Proposed Answer: {}. "
        default_template = qa_text + self.instruction
        return default_template

    def _extract_answers(self, responses: List[str]) -> List[float]:
        """
        List-level implementation of _extract_single_answer
        """
        return [self._extract_single_answer(r) for r in responses]

    def _extract_single_answer(self, response: str) -> float:
        """
        A method to extract score from an llm response based on provided score-keyword dictionary.
        """
        if self.scoring_template == "continuous":
            score = "".join(c for c in response if c.isdigit())
            if len(score) > 0:
                if 0.0 <= float(score) <= 100.0:
                    return float(score) / 100.0  # normalize
        elif self.scoring_template in ["true_false_uncertain", "true_false", None]:
            response = response.lower()
            for score, keywords in self.keywords_to_scores_dict.items():
                if any(keyword in response for keyword in keywords):
                    return score
        return np.nan

    def _validate_inputs(self):
        """Validate inputs"""
        if self.template_ques_ans and self.keywords_to_scores_dict:
            for key, val in self.keywords_to_scores_dict.items():
                if not isinstance(key, float):
                    raise ValueError("keys in keywords_to_scores_dict must be floats")
                if not isinstance(val, list):
                    raise ValueError(
                        "values in keywords_to_scores_dict must be lists of strings"
                    )
                # TODO: validate value ordering for substrings of other keys
        if self.scoring_template in TEMPLATE_TO_INSTRUCTION:
            self.instruction = TEMPLATE_TO_INSTRUCTION[self.scoring_template]
            self.template_ques_ans = self._default_template_ques_ans()
            self.keywords_to_scores_dict = {
                round(k, 1): v for k, v in KEYWORDS_TO_SCORES_DICT.items()
            }
            if self.scoring_template == "true_false":  # drop uncertain option if binary
                del self.keywords_to_scores_dict[0.5]
        else:
            raise ValueError(
                """If provided, scoring_template must be one of 'true_false_uncertain', 'true_false', 'continuous'. Otherwise, valid template_ques_ans and keywords_to_scores_dict must be provided"""
            )